# -*- coding: utf-8 -*-

class QC:
    FAIL = 'FAIL'
    WARNING = 'WARNING'
    PASS = 'PASS'
    # Errors for Hansel
    MISSING_TILES_ERROR_1 = 'Missing Tiles Error 1'
    MIXED_SAMPLE_ERROR_2 = 'Mixed Sample Error 2'
    AMBIGUOUS_RESULTS_ERROR_3 = 'Ambiguous Results Error 3'
    UNCONFIDENT_RESULTS_ERROR_4 = 'Unconfident Results Error 4'
    INTERMEDIATE_SUBTYPE_WARNING = 'Intermediate Subtype'
    LOW_COVERAGE_WARNING = 'Low Coverage'
    NO_SUBTYPE_RESULT = 'No subtype result!'
